# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


from lot_basis import *

from molmod import angstrom, Molecule
from molmod.periodic import periodic
from molmod.io import FCHKFile

import os, numpy


__all__ = ["get_root", "states"]


def get_root(lot_label, basis_label, suffix):
    result = "%s__%s" % (lot_label, basis_label)
    if len(suffix) > 0:
        result = "%s__%s" % (result, suffix)
    return result


class State(object):
    def __init__(self, name, jobs):
        self.name = name
        self.jobs = jobs


class G03Job(object):
    def __init__(self, name, cmd, post=""):
        self.name = name
        self.cmd = cmd
        if isinstance(post, list):
            self.post = "\n".join(post)
        else:
            self.post = post

    def get_mol(self, state, root):
        if self.name == "opt":
            # load the initial geometry
            mol = Molecule.from_file("init/%s.xyz" % state.name)
        else:
            # load the optimized geometry
            fn_fchk = "%s/%s__opt/gaussian.fchk" % (root, state.name)
            mol = FCHKFile(fn_fchk, field_labels=[]).molecule
        f = file("init/%s.fragments" % state.name)
        mol.charge_mult = f.readline().split()
        mol.tags = f.readline().split()
        f.close()
        return mol

    def write_input(self, state, root, lot_label, basis_label, suffix="", random=False):
        try:
            mol = self.get_mol(state, root)
        except IOError:
            return

        if self.name.startswith("cps"):
            i = int(self.name[-1])
            mult = int(mol.charge_mult[2*i+1])
        else:
            mult = int(mol.charge_mult[1])
        if lot_label.startswith("ro"):
            lot = get_lot(lot_label[2:], mult, restricted=True)
        else:
            lot = get_lot(lot_label, mult)
        basis = get_basis(basis_label)

        dirname = "%s/%s__%s" % (root, state.name, self.name)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)

        fn_com = "%s/gaussian.com" % dirname
        if os.path.isfile(fn_com):
            return

        # copy an initial guess of the wavefunction
        destination = "%s/gaussian.in.fchk" % dirname
        if not (self.name == "opt" or self.name == "bsse" or self.name.startswith("cps")):
            if os.path.isfile(destination):
                os.remove(destination)
            os.symlink(
                "../%s__opt/gaussian.fchk" % state.name,
                destination,
            )

        # write the input file
        f = file(fn_com, "w")
        print >> f, "%chk=gaussian.chk"
        print >> f, "%nproc=1"
        print >> f, "%mem=1000MB"
        print >> f, "#p %s/%s %s %s maxdisk=5GB" % (lot.name, basis.name, self.cmd, lot.iop),
        if basis.diffuse:
            print >> f, "scf=tight",
        if not (self.name == "opt" or self.name == "bsse" or self.name.startswith("cps")):
            print >> f, "guess=read",
        if len(lot.extra_overlay) > 0:
            print >> f, "extraoverlay",
        print >> f
        print >> f
        if len(lot.extra_overlay) > 0:
            print >> f, lot.extra_overlay
            print >> f
        print >> f, dirname
        print >> f
        if self.name == "bsse":
            print >> f, " ".join(mol.charge_mult)
        elif self.name.startswith("cps"):
            i = int(self.name[-1])
            print >> f, " ".join(mol.charge_mult[2*i:2*i+2])
        else:
            print >> f, " ".join(mol.charge_mult[:2])
        for i in xrange(mol.size):
            if random and self.name == "opt":
                x, y, z = mol.coordinates[i]/angstrom + numpy.random.uniform(-0.1,0.1,3)
            else:
                x, y, z = mol.coordinates[i]/angstrom
            symbol = periodic[mol.numbers[i]].symbol
            if self.name == "bsse":
                tag = mol.tags[i]
            else:
                tag = ""
            if self.name.startswith("cps"):
                if self.name.startswith("cps_full"):
                    if self.name[-1] != mol.tags[i] and self.name[-1] != "0":
                        symbol += "-Bq"
                else:
                    if self.name[-1] != mol.tags[i]:
                        continue
            print >> f, " %2s   % 10.5f   % 10.5f   % 10.5f  %s" % (
                symbol, x, y, z, tag
            )
        print >> f
        print >> f, self.post
        print >> f
        print >> f
        f.close()

        print dirname




states = [
    State("ethene", [
        G03Job("opt", "opt"),
        G03Job("freq", "freq(noraman)"),
        G03Job("sp", "sp"),
    ]),

    State("ethyl", [
        G03Job("opt", "opt"),
        G03Job("freq", "freq(noraman)"),
        G03Job("sp", "sp"),
        G03Job("scan_methyl", "opt(modredundant)", "4 1 2 6 S 72 5.0"),
    ]),

    State("ts_ad1_trans", [
        G03Job("opt", "opt(ts,calcfc,noeigentest)"),
        G03Job("freq", "freq(noraman)"),
        G03Job("bsse", "counterpoise=2"),
        G03Job("cps_full_0", "sp"),
        G03Job("cps_full_1", "sp"),
        G03Job("cps_full_2", "sp"),
        G03Job("cps_sole_1", "sp"),
        G03Job("cps_sole_2", "sp"),
        G03Job("sp", "sp"),
        G03Job("scan_forming_bond", "opt(ts,calcfc,modredundant,noeigentest)", ["1 2 3 4 S 72 5.0"]),
        G03Job("scan_methyl", "opt(ts,calcfc,modredundant,noeigentest)", ["5 1 2 8 S 72 5.0", "1 2 3 5 B"]),
    ]),

    State("ts_ad1_gauche", [
        G03Job("opt", "opt(ts,calcfc,noeigentest)"),
        G03Job("freq", "freq(noraman)"),
        G03Job("bsse", "counterpoise=2"),
        G03Job("cps_full_0", "sp"),
        G03Job("cps_full_1", "sp"),
        G03Job("cps_full_2", "sp"),
        G03Job("cps_sole_1", "sp"),
        G03Job("cps_sole_2", "sp"),
        G03Job("sp", "sp"),
        G03Job("scan_methyl", "opt(ts,calcfc,modredundant,noeigentest)", ["5 1 2 8 S 72 5.0", "1 2 3 5 B"]),
    ]),
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


# Import libraries.
from tamkin import *   # The TAMkin library

# Load the molecules (including the Hessian etc.)
mol_oxygen = load_molecule_g03fchk("oxygen.fchk")
mol_hydrogen = load_molecule_g03fchk("hydrogen.fchk")
mol_water = load_molecule_g03fchk("water.fchk")
# Perform normal mode analysis on the molecules
nma_oxygen = NMA(mol_oxygen, ConstrainExt())
nma_hydrogen = NMA(mol_hydrogen, ConstrainExt())
nma_water = NMA(mol_water, ConstrainExt())
# Construct the partition functions.
pf_oxygen = PartFun(nma_oxygen, [ExtTrans(), ExtRot()])
pf_hydrogen = PartFun(nma_hydrogen, [ExtTrans(), ExtRot()])
pf_water = PartFun(nma_water, [ExtTrans(), ExtRot()])
# Define a kinetic model for the chemical reaction.
tm = ThermodynamicModel([pf_oxygen, (pf_hydrogen, 2)], [(pf_water, 2)])
# Write tables with the principal energies at 298.15K
tm.write_table(298.15, "formation.csv")
# Write an overview of the thermodynamic model to a file
tm.write_to_file("formation.txt")



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


from tamkin.data import Molecule

from molmod import angstrom, amu, calorie, avogadro, lightspeed, centimeter
from molmod.periodic import periodic

import numpy as np


__all__ = [
    "load_molecule_charmm", "load_coordinates_charmm", "load_modes_charmm",
    "load_peptide_info_charmm",
]


def load_molecule_charmm(charmmfile_cor, charmmfile_hess, is_periodic=False):
    """Read from Hessian-CHARMM-file format

       Arguments:
        | charmmfile_cor  --  the filename of the .cor file
        | charmmfile_hess  --  the filename of the Hessian file

       Optional argument:
        | is_periodic  --  True when the system is periodic in three dimensions.
                           False when the systen is aperiodic. [default=True]
    """
    f = file(charmmfile_hess)
    # skip lines if they start with a *
    while True:
        line = f.readline()
        if not line.startswith("*"): break
    N = int(line.split()[-1])
    assert N > 0   # nb of atoms should be > 0

    energy = float(f.readline().split()[-1]) * 1000*calorie/avogadro

    gradient = np.zeros((N,3),float)
    for i,line in enumerate(f):
        words = line.split()
        gradient[i,:] = [float(word) for word in words]
        if i == (N-1):
            break
    gradient *= 1000*calorie/avogadro/angstrom
    hessian = np.zeros((3*N,3*N),float)
    row = 0
    col = 0
    for line in f:
        element = float(line.split()[-1])
        hessian[row,col]=element
        hessian[col,row]=element
        col += 1
        if col>=3*N:        #if this new col doesn't exist
            row += 1        #go to next row
            col = row       #to diagonal element
            if row >= 3*N:  #if this new row doesn't exist
               break
    hessian = hessian * 1000*calorie/avogadro /angstrom**2

    positions = np.zeros((N,3),float)
    for i,line in enumerate(f):
        words = line.split()
        positions[i,:] = [float(word)*angstrom for word in words]
        if i == (N-1):
            break
    f.close()

    # Read from coordinates-CHARMM-file
    # format:  header lines, which start with *
    #          N lines with   - mass in last column
    #                         - atomic type in 4th column
    f = file(charmmfile_cor)
    masses = np.zeros(N,float)
    symbols  = []
    for line in f:
        if not line.startswith("*"): # skip header lines
            break
    for i,line in enumerate(f):
        words = line.split()
        masses[i] = float(words[-1])*amu   # mass
        symbols.append( words[3] )         # symbol
        if i == (N-1):
            break
    f.close()

    # get corresponding atomic numbers
    mass_table = np.zeros(len(periodic))
    for i in xrange(1, len(mass_table)):
        m1 = periodic[i].mass
        if m1 is None:
            m1 = 200000.0
        m2 = periodic[i+1].mass
        if m2 is None:
            m2 = 200000.0
        mass_table[i] = 0.5*(m1+m2)
    atomicnumbers = np.zeros(N, int)
    for i,mass in enumerate(masses):
        atomicnumbers[i] = mass_table.searchsorted(mass)

    return Molecule(
        atomicnumbers, positions, masses, energy, gradient, hessian,
        1, # multiplicity
        symmetry_number=1, periodic=is_periodic, symbols=tuple(symbols)
    )


def load_coordinates_charmm(filename):
    """Read coordinates from a standard CHARMM coordinate file

       Arguments:
         | filename  --  the CHARMM coordinate file (typically extension .crd or
                       .cor)

       Return values:
         | coordinates  --  coordinates in atomic units, numpy array with
                            shape (N,3)
         | masses  --  atomic masses in atomic units
         | symbols  --  list of CHARMM atom symbols
    """

    # skip the lines that start with * comments
    f = open(filename,'r')
    for line in f:
        if not line.startswith("*"): break
    N = int(line.split()[0])   # nb of atoms

    # store coordinates in Nbx3 matrix
    symbols = ['']*N
    coordinates = np.zeros((N,3),float)
    masses = np.zeros(N,float)
    count = 0
    for line in f:
        words = line.split()
        symbols[count]       = words[3]
        coordinates[count,:] = np.array([float(word) for word in words[4:7]])*angstrom
        masses[count]        = float(words[9])*amu
        count += 1
        if count >= N: break
    f.close()
    return coordinates, masses, symbols


def load_modes_charmm(filename):
    """Read modes, frequencies and masses from a standard CHARMM-modes-file

       The CHARMM-modes-file is generated by the VIBRAN module in CHARMM.

       Argument:
         | filename  --  CHARMM-modes-file

       Return values:
         | modes  --  numpy array with mass-weighted modes in the columns
         | freqs  --  numpy array with frequencies in atomic units
         | masses  --  atomic masses in atomic units
    """
    f = file(filename)

    # skip the lines that start with * comments
    for line in f:
        if not line.strip().startswith("*"): break

    # read nb of atoms and nbfreqs (if not yet specified by user)
    words = line.split()        # the current line does not start with a *
    nbfreqs = int(words[0])
    N       = int(words[1])/3   # nb of atoms

    # lines with masses, 6 masses on each line
    nblines = int(np.ceil(N/6.0))
    masses = np.zeros(N,float)
    count = 0
    for line in f:
        words = line.split()
        n = len(words)
        masses[count:count+n] = np.array([float(word) for word in words])
        count += n
        if count >= N: break

    # read nbfreqs freqs
    CNVFRQ = 2045.5/(2.99793*6.28319)  # conversion factor, see c36a0/source/fcm/consta.fcm in charmm code
    nblines = int(np.ceil(nbfreqs/6.0))
    freqs = np.zeros(nbfreqs, float)
    countline = 0
    countfreq = 0
    for line in f:
        words = line.split()
        for word in words:
            # do conversion
            freq_sq = float(word) #squared value
            if freq_sq > 0.0:  freq =  np.sqrt( freq_sq)
            else:              freq = -np.sqrt(-freq_sq) #actually imaginary
            freqs[countfreq] = freq * CNVFRQ * lightspeed/centimeter # conversion factor CHARMM, put into Tamkin internal units
            countfreq += 1
        countline += 1
        if countline >= nblines: break
    if countfreq != nbfreqs:
        raise ValueError("should have read "+str(nbfreqs)+" frequencies, but read "+str(countfreq))

    # read the nbfreqs modes
    modes = np.zeros((3*N,nbfreqs),float)
    row = 0
    col = 0
    for line in f:
        words = line.split()
        n = len(words)
        modes[row:row+n,col] = np.array([float(word) for word in words])
        row += n
        if row == 3*N:
            col += 1
            row = 0

    f.close()
    return modes, freqs, masses


def load_peptide_info_charmm(filename):
    """Load information from CHARMM file for peptide blocks and subsystems

       Arguments:
         | filename  --  the CHARMM coordinate file (typically extension .crd
                         or .cor)

       Return values:
         | N  --  total number of atoms in the peptide
         | calpha  --  indices of the alpha carbons ('CA' in CHARMM file)
         | proline  --  indices of the alpha carbons that belong to proline
                        residues ('PRO  CA' in CHARMM file)
         | carbon  -- indices of the backbone carbons, exclude the alpha carbons
                      ('C' in CHARMM file)
         | oxygen  -- indices of the backbone oxygens ('O' in CHARMM file)
         | nitrogen  --  indices of the backbone nitrogens ('N' in CHARMM file)
    """
    # Reading from charmmfile
    f = file(filename)
    # nb of atoms
    for i,line in enumerate(f):
        words = line.split()
        if words[0]!="*":
            N = int(words[0])
            break
    # find alpha carbons, proline residues, carbons, oxygens, nitrogens
    calpha = []
    proline = []
    carbon = []
    oxygen = []
    nitrogen = []
    for i,line in enumerate(f):
        words = line.split()
        if words[3].startswith("CA"):
            calpha.append(int(words[0]))
            if words[2]=="PRO":
                proline.append(int(words[0]))
        if words[3]=="C":
            carbon.append(int(words[0]))
        if words[3]=="O":
            oxygen.append(int(words[0]))
        if words[3]=="N":
            nitrogen.append(int(words[0]))
    f.close()
    return N, calpha, proline, carbon, oxygen, nitrogen



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


# Import the tamkin libarary.
from tamkin import *
# Import unit conversin factors
from molmod import *
# Import numpy stuff
from numpy import *

###
### reaction can be split up into two parts (HAA = acetic acid):
### activation part: VO(AA)(OH)--H2O + TBHP <=> VO(AA)(OOtBu) + 2H2O
### reaction part (oxygen transfer reaction): VO(AA)(OOtBu) + cyclohexene -> VO(ACAC)(OtBu) + cyclohexene-epoxide
###
# yclohexene  reaction.py  TBHP  TS  VO_AA_OOtBu  water

nma_H2O = NMA(load_molecule_g03fchk("water/gaussian.fchk"))
nma_VO_AA_OH_H2O = NMA(load_molecule_g03fchk("VO_AA_OH_H2O/gaussian.fchk"))
nma_VO_AA_OOtBu = NMA(load_molecule_g03fchk("VO_AA_OOtBu/gaussian.fchk"))
nma_TBHP = NMA(load_molecule_g03fchk("TBHP/gaussian.fchk"))
nma_TS = NMA(load_molecule_g03fchk("TS/TS.fchk"))
nma_cyclohexene = NMA(load_molecule_g03fchk("cyclohexene/gaussian.fchk"))

pf_H2O = PartFun(nma_H2O, [ExtTrans(), ExtRot(2),Vibrations(freq_scaling=1.0, zp_scaling=1.0)])
pf_VO_AA_OH_H2O = PartFun(nma_VO_AA_OH_H2O, [ExtTrans(), ExtRot(1),Vibrations(freq_scaling=1.0, zp_scaling=1.0)])
pf_VO_AA_OOtBu = PartFun(nma_VO_AA_OOtBu, [ExtTrans(), ExtRot(1),Vibrations(freq_scaling=1.0, zp_scaling=1.0)])
pf_TBHP = PartFun(nma_TBHP, [ExtTrans(), ExtRot(1),Vibrations(freq_scaling=1.0, zp_scaling=1.0)])
pf_TS =  PartFun(nma_TS, [ExtTrans(), ExtRot(1),Vibrations(freq_scaling=1.0, zp_scaling=1.0)])
pf_cyclohexene = PartFun(nma_cyclohexene, [ExtTrans(), ExtRot(1),Vibrations(freq_scaling=1.0, zp_scaling=1.0)])

# One can compute the equilibrium coefficient for the activation reaction and the forward coefficient for the reaction separately:

# Thermodynamic model
tm = ThermodynamicModel([pf_VO_AA_OH_H2O,pf_TBHP],[pf_VO_AA_OOtBu,pf_H2O,pf_H2O])
print "Thermodynamic model:"
print "equilibrium constant and gibbs free energy difference"
print tm.equilibrium_constant(323.0,do_log=False)/tm.unit, tm.unit_name,"    " ,tm.free_energy_change(323.0)/kjmol, "kJ/mol"
print
# Kinetic model
km = KineticModel([pf_VO_AA_OOtBu,pf_cyclohexene],pf_TS,tunneling=None)
print "Kinetic model:"
print "rate constant and gibbs free energy difference"
print km.rate_constant(323.0,do_log=False)/km.unit, "%s" % km.unit_name,"   " , km.free_energy_change(323.0)/kjmol, "kJ/mol"
print
ra = ReactionAnalysis(km, 273, 373.1, temp_step=10)
ra.plot_arrhenius("arrhenius.png") # make the Arrhenius plot

# Estimate the error on the kinetic parameters due to level of theory artifacts
# with Monte Carlo sampling. The monte_carlo method takes three optional
# arguments:
#  1) freq_error: the absolute stochastic error on the frequencies (default=1*invcm)
#  2) energy_error: the absolute error on the energy (default=0.0)
#  3) num_iter: the number of monte carlo samples (default=100)
ra.monte_carlo()
# plot the parameters, this includes the monte carlo results
ra.plot_parameters("parameters.png")
# write all results to a file.
ra.write_to_file("reaction.txt") # summary of the analysis


# Or one can determine a global k value and fit this within a temperature region by making use of the class ActivationKineticModel
akm = ActivationKineticModel(tm,km)
rakm = ReactionAnalysis(akm,273,373.1,temp_step=10)
rakm.plot_arrhenius("arrhenius_akm.png")
rakm.monte_carlo()
rakm.plot_parameters("parameters_akm.png")
print "Activation kinetic model:"
print "global rate constant and global gibbs free energy difference"
print akm.rate_constant(323.0)/akm.unit, "%s" % akm.unit_name,"    ", akm.free_energy_change(323.0)/kjmol, "kJ/mol"
akm.write_to_file("activation_model.txt")



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --

from tamkin.data import Molecule

from molmod import angstrom, amu, calorie, avogadro
from molmod.periodic import periodic

import numpy as np


__all__ = ["load_molecule_qchem"]

def load_molecule_qchem(qchemfile, hessfile = None, multiplicity=1, is_periodic = False):
    """Load a molecule from a Q-Chem frequency run

       Arguments:
        | qchemfile  --  Filename of the Q-Chem computation output.

       Optional arguments:
        | hessfile  --  Filename of a separate Hessian file.
        | multiplicity  --  The spin multiplicity of the electronic system
                            [default=1]
        | is_periodic  --  True when the system is periodic in three dimensions.
                           False when the systen is nonperiodic. [default=False]

       Whether the Hessian is printed to a separate Hessian file, depends on the
       used version of Q-Chem. The use of the separate Hessian file is slightly
       more accurate, because the number of printed digits is higher than in the
       Q-Chem output file.

       **Warning**

       At present, the gradient is set to a Nx3 array of zero values, since the
       gradient is not printed out in the Q-Chem output file in general. This
       means that the value of the gradient should be checked before applying
       methods designed for partially optimized structures (currently PHVA,
       MBH and PHVA_MBH).
    """
    # TODO fill in keyword for printing hessian
    f = file(qchemfile)
    # get coords
    for line in f:
        if line.strip().startswith("Standard Nuclear Orientation (Angstroms)"):
            break
    f.next()
    f.next()
    positions = []
    symbols = []
    for line in f:
        if line.strip().startswith("----"): break
        words = line.split()
        symbols.append(words[1])
        coor = [float(words[2]),float(words[3]),float(words[4])]
        positions.append(coor)
    positions = np.array(positions)*angstrom
    N = len(positions)    #nb of atoms

    numbers = np.zeros(N,int)
    for i, symbol in enumerate(symbols):
        numbers[i] = periodic[symbol].number
    #masses = np.zeros(N,float)
    #for i, symbol in enumerate(symbols):
    #    masses[i] = periodic[symbol].mass

    # grep the SCF energy
    energy = None
    for line in f:
        if line.strip().startswith("Cycle       Energy         DIIS Error"):
            break
    for line in f:
        if line.strip().endswith("met"):
            energy = float(line.split()[1]) # in hartree
            break

    # get Hessian
    hessian = np.zeros((3*N,3*N),float)
    if hessfile is None:
      for line in f:
          if line.strip().startswith("Hessian of the SCF Energy") or line.strip().startswith("Final Hessian"):
              break
      nb = int(np.ceil(N*3/6))
      for i in range(nb):
          f.next()
          row = 0
          for line in f:
              words = line.split()
              hessian[row, 6*i:6*(i+1)] = np.array(sum([[float(word)] for word in words[1:]],[])) #/ angstrom**2
              row += 1
              if row >= 3*N : break

    # get masses
    masses = np.zeros(N,float)
    for line in f:
        if line.strip().startswith("Zero point vibrational"):
            break
    f.next()
    count=0
    for line in f:
        masses[count] = float(line.split()[-1])*amu
        count += 1
        if count >= N : break

    # get Symm Nb
    for line in f:
        if line.strip().startswith("Rotational Symmetry Number is"):
            break
    symmetry_number = int(line.split()[-1])
    f.close()

    # or get Hessian from other file
    if hessfile is not None:
      f = file(hessfile)
      row = 0
      col = 0
      for line in f:
          hessian[row,col] = float(line.split()[0]) *1000*calorie/avogadro /angstrom**2
          col += 1
          if col >= 3*N:
              row += 1
              col = row
      f.close()
      for i in range(len(hessian)):
          for j in range(0,i):
              hessian[i,j] = hessian[j,i]

    # get gradient   TODO
    gradient = np.zeros((N,3), float)

    return Molecule(
        numbers, positions, masses, energy, gradient, hessian, multiplicity,
        symmetry_number, is_periodic
    )



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Analysis of molecular geometries"""

import numpy as np


__all__ = ["transrot_basis", "rank_linearity"]


def transrot_basis(coordinates, rot=True):
    """Constructs 6 vectors which represent global translations/rotations.

       Argument:
        | coordinates  --  The atom coordinates (float numpy array with shape
                           Nx3)

       Optional argument:
        | rot  --  When True the rotations are included [default=True]

       The return value is a numpy array with 3*N columns and 6 (``rot==True``)
       or 3 (``rot==False``) rows. The rows are not mass weighted.
    """
    if not rot:
        result = np.zeros((3, coordinates.size), float)
    else:
        result = np.zeros((6, coordinates.size), float)
    # translation
    result[0, 0::3] = 1
    result[1, 1::3] = 1
    result[2, 2::3] = 1
    if rot:
        # rotation
        result[3, 1::3] =  coordinates[:,2]
        result[3, 2::3] = -coordinates[:,1]
        result[4, 2::3] =  coordinates[:,0]
        result[4, 0::3] = -coordinates[:,2]
        result[5, 0::3] =  coordinates[:,1]
        result[5, 1::3] = -coordinates[:,0]
    return result


def rank_linearity(coordinates, svd_threshold=1e-5, masses3=None):
    """Test the linearity of the given coordinates

       Arguments:
        | coordinates  --  The atom coordinates (float numpy array with shape
                           Nx3)

       Optional argument:
        | svd_threshold  --  Defines the sensitivity for deviations from
                             linearity [default=1e-5]
        | masses3  --  array with atomic masses (each mass is repeated trice)

       Returns the number of external degrees of freedom (dof) and the
       corresponding basis. The dof can be

       * 6 dof if atoms of system are non-collinear
       * 5 dof if atoms of system are collinear, e.g. when the subsystem
         contains only 2 atoms
       * 3 dof if system contains just 1 atom

       If masses3 is given, the analysis is performed in mass-weighted
       coordinates and the returned basis is in mass-weighted coordinates.
       If masses3 is not given everything is done in non-mass-weighted Cartesian
       coordinates.
    """
    if masses3 is None:
        center = coordinates.mean(0)  # geometrical center
        transrot = transrot_basis(coordinates - center)
    else:
        center = (coordinates*masses3.reshape((-1,3))).sum(0)/(masses3.sum()/3)  # center of mass
        #print ((coordinates-center)*masses3.reshape((-1,3))).sum(0)  # check
        transrot = transrot_basis(coordinates - center)
        transrot *= np.sqrt(masses3)

    U, W, Vt = np.linalg.svd(transrot, full_matrices=False)
    rank = (abs(W) > abs(W[0])*svd_threshold).sum()
    external_basis = Vt[:rank]

    return rank, external_basis



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


# Import the tamkin libarary.
from tamkin import *
# Import unit conversin factors
from molmod import *
# Import numpy stuff
from numpy import *


# Load the gaussian data.
molecule = load_molecule_g03fchk("freq/gaussian.fchk")
print "Energy [kJ/mol]:", molecule.energy/kjmol
rot_scan = load_rotscan_g03log("scan/gaussian.log")


# A) Perform a standard normal mode analysis
nma = NMA(molecule, ConstrainExt())
rotor = Rotor(rot_scan, molecule, rotsym=1, even=True)
pf = PartFun(nma, [ExtTrans(), ExtRot(), rotor])
print "Heat capacity at 300K, constant pressure [J/(mol*K)]:", pf.heat_capacity(300*kelvin)/(joule/mol/kelvin)
# Write some general information about the molecule and the partition function
# to a file.
pf.write_to_file("partfun.txt")
rotor.plot_levels("energy_levels.png", 300, do_levels=False, do_data=False)

I = rotor.reduced_moment
for angle, energy in array(rot_scan.potential).transpose():
    K = rotor.hb.eval_deriv2(array([angle]), rotor.v_coeffs)[0]
    nu = sqrt(abs(K/I))
    hertz = 1/second
    print "Angle in deg:", angle/deg
    print "Energy in kJ/mol", energy/joule
    print "Frequency in THz:", nu/hertz/1e12/(2*pi)
    print "Wavenumber in cm-1:", nu/(lightspeed/centimeter)/(2*pi)
    print



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


# Import the tamkin library.
from tamkin import *
from molmod import * # for the units
from molmod.isotopes import ame2003


def create_kinetic_model(mol_react, mol_trans):
    # Perform normal mode analysis on the three molecules
    nma_react = NMA(mol_react, ConstrainExt())
    nma_trans = NMA(mol_trans, ConstrainExt())
    # Construct the two partition functions.
    pf_react = PartFun(nma_react, [ExtTrans(), ExtRot(1), Vibrations(classical=True, freq_scaling=0.9085)])
    pf_trans = PartFun(nma_trans, [ExtTrans(), ExtRot(1), Vibrations(classical=True, freq_scaling=0.9085)])

    return KineticModel([pf_react], pf_trans)


old_mol_react = load_molecule_g03fchk("reactant.fchk")
old_mol_trans = load_molecule_g03fchk("trans.fchk")
old_km = create_kinetic_model(old_mol_react, old_mol_trans)
print "Original rate constant at 303K =", old_km.rate_constant(303)/old_km.unit, old_km.unit_name

# does not work:
# mol_react1.masses[0] = 13*amu

new_masses = old_mol_react.masses.copy()
new_masses[20] = ame2003.masses[7][15]
new_mol_react = old_mol_react.copy_with(masses=new_masses)
new_mol_trans = old_mol_trans.copy_with(masses=new_masses)
new_km = create_kinetic_model(new_mol_react, new_mol_trans)
print "New rate constant at 303K =", new_km.rate_constant(303)/new_km.unit, new_km.unit_name

print "Ratio at 303K =", old_km.rate_constant(303)/new_km.rate_constant(303)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--

from lot_basis import lots_list
import html

from molmod.io.gaussian03.fchk import FCHKFile
from molmod.units import angstrom, deg
import molmod.ic as ic

def bond_length(mol, i0, i1):
    c = mol.coordinates
    return ic.bond_length(c[i0], c[i1])[0]

def bend_angle(mol, i0, i1, i2):
    c = mol.coordinates
    return ic.bend_angle(c[i0], c[i1], c[i2])[0]


def generic_geom(ic_descs, fn_fchk, name, basis_label, rows):
    if len(rows) == 0:
        rows.append([u"<th>Approach→<br />Functional↓</th>"])
        rows.append(["<th></th>"])
    rows[0].append("<th colspan=%i>%s<br />%s</th>" % (len(ic_descs), name, basis_label))
    for ic_desc in ic_descs:
        rows[1].append(ic_desc[0])
    counter = 2
    for lot in lots_list:
        if lot.spin == "ROS":
            lot_label = "ro" + lot.label
        else:
            lot_label = lot.label
        if len(rows) <= counter:
            rows.append(["<th>%s</th>" % lot_label])
        try:
            mol = FCHKFile(fn_fchk % (lot_label, basis_label), field_labels=[]).molecule
        except IOError:
            mol = None
        if mol is None:
            for ic_desc in ic_descs:
                rows[counter].append("<td>&nbsp</td>")
        else:
            for ic_desc in ic_descs:
                cell_format, ic_fn, indexes, unit = ic_desc[1:]
                rows[counter].append(cell_format % (ic_fn(mol, *indexes)/unit))
        counter += 1

def ts_geom(ts_name, basis_label, rows):
    fn_fchk = "%%s__%%s/ts_ad1_%s__opt/gaussian.fchk" % ts_name.lower()
    ic_descs = [
        (u"<td>d<sub>12</sub>/Å</td>", "<td style='background-color:#DDD;'>%.2f</td>", bond_length, (1, 2), angstrom),
        (u"<td>a<sub>012</sub>/°</td>", "<td>%.1f</td>", bend_angle, (0, 1, 2), deg),
        (u"<td>a<sub>123</sub>/°</td>", "<td>%.1f</td>", bend_angle, (1, 2, 3), deg),
    ]
    generic_geom(ic_descs, fn_fchk, ts_name, basis_label, rows)

def ra_geom(mol_name, basis_label, rows):
    fn_fchk = "%%s__%%s/%s__opt/gaussian.fchk" % mol_name.lower()
    ic_descs = [
        (u"<td>d<sub>01</sub>/Å</td>", "<td>%.2f</td>", bond_length, (0, 1), angstrom),
    ]
    generic_geom(ic_descs, fn_fchk, mol_name, basis_label, rows)


f = file("geotab.html", "w")
print >> f, html.header % "GEO Overview"

rows = []
ts_geom("Gauche", "6-31gd", rows)
ts_geom("Trans", "6-31gd", rows)
ts_geom("Gauche", "6-311+g3df2p", rows)
ts_geom("Trans", "6-311+g3df2p", rows)

print >> f, "<p>Geometrical parameters related to the transition state.</p>"
html.print_table(f, rows)

rows = []
ra_geom("Ethene", "6-31gd", rows)
ra_geom("Ethyl", "6-31gd", rows)
ra_geom("Ethene", "6-311+g3df2p", rows)
ra_geom("Ethyl", "6-311+g3df2p", rows)

print >> f, "<p>Geometrical parameters related to the reactants.</p>"
html.print_table(f, rows)

print >> f, html.footer



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Data containers for input from QM or MM simulation codes

   Most IO routines in :mod:`tamkin.io` return instances of the classes
   defined here. These objects are just read-only containers for the QM or MM
   output with a standardized interface.
"""

from tamkin.geom import transrot_basis

from molmod import Molecule as BaseMolecule, MolecularGraph, ReadOnly, \
    UnitCell, ReadOnlyAttribute
from molmod.periodic import periodic
from molmod.graphs import cached

import numpy as np


__all__ = ["Molecule", "BareNucleus", "Proton", "RotScan",
           "translate_pbc"]


class Molecule(BaseMolecule):
    """A container for a Hessian computation output from QM or MM codes."""

    def check_gradient(self, gradient):
        if len(gradient) != self.size:
            raise TypeError("The size of the gradient does not match the "
                "length of the atomic numbers array.")

    def check_hessian(self, hessian):
        if hessian.shape != (self.size*3, self.size*3):
            raise TypeError("The Hessian must be a 3N by 3N matrix where N is "
                "the number of atoms.")

    energy = ReadOnlyAttribute(float, none=False)
    gradient = ReadOnlyAttribute(np.ndarray, none=False,
        check=check_gradient, npdim=2, npshape=(None, 3), npdtype=float)
    hessian = ReadOnlyAttribute(np.ndarray, none=False,
        check=check_hessian, npdim=2, npdtype=float)
    multiplicity = ReadOnlyAttribute(int)
    symmetry_number = ReadOnlyAttribute(int)
    periodic = ReadOnlyAttribute(bool)
    fixed = ReadOnlyAttribute(np.ndarray, npdim=1, npdtype=int)

    def __init__(self, numbers, coordinates, masses, energy, gradient, hessian, multiplicity=None, symmetry_number=None, periodic=False, title=None, graph=None, symbols=None, unit_cell=None, fixed=None):
        """
           Arguments:
            | ``numbers`` -- The atom numbers (integer numpy array with shape N)
            | ``coordinates`` -- the atom coordinates in Bohr (float numpy array
                                 with shape Nx3)
            | ``masses`` -- The atomic masses in atomic units (float numpy array
                            with shape N)
            | ``energy`` -- The molecular energy in Hartree
            | ``gradient`` -- The gradient of the energy, i.e. the derivatives
                              towards Cartesian coordinates, in atomic units
                              (float numpy array with shape Nx3)
            | ``hessian`` -- The hessian of the energy, i.e. the matrix with
                             second order derivatives towards Cartesian
                             coordinates, in atomic units (float numpy array
                             with shape 3Nx3N)
            | ``multiplicity`` -- The spin multiplicity of the electronic system

           Optional arguments:
            | ``symmetry_number`` -- The rotational symmetry number, None when
                                     not known or computed [default=None]
            | ``periodic`` -- True when the system is periodic in three
                              dimensions [default=False]
            | ``title`` -- The title of the system
            | ``graph`` -- The molecular graph of the system
            | ``symbols`` -- A list with atom symbols
            | ``unit_cell`` -- The unit cell vectors for periodic structures
            | ``fixed`` -- An array with indices of fixed atoms
        """
        BaseMolecule.__init__(self, numbers, coordinates, title, masses, graph, symbols, unit_cell)
        self.energy = energy
        self.gradient = gradient
        self.hessian = hessian
        self.multiplicity = multiplicity
        self.symmetry_number = symmetry_number
        self.periodic = periodic
        self.fixed = fixed

    def get_external_basis_new(self, im_threshold=1.0):
        """Create a robust basis for small displacements in the external degrees of freedom.

           The basis is expressed in mass-weighted Cartesian coordinates.
           The three translations are along the x, y, and z-axis. The three
           rotations are about an axis parallel to the x, y, and z-axis
           through the center of mass.

           The returned result depends on the periodicity of the system and on
           the parameter im_threshold:

           * When the system is periodic, only the translation external degrees
             are included. The result is an array with shape (3,3N)

           * When the system is not periodic, the rotational external
             degrees are also included. The result is an array with shape (5,3N)
             or (6,3N). The first three rows correspond to translation, the
             latter rows correspond to rotation. Rotational modes that have a
             angular moment below ``im_threshold`` are discarded.
        """
        if self.periodic:
            result = np.zeros((3, 3*self.size), float)
        else:
            ims, iaxs = np.linalg.eigh(self.inertia_tensor)
            mask = ims > im_threshold
            ims = ims[mask]
            iaxs = iaxs[:,mask]
            ext_dof = len(ims) + 3
            result = np.zeros((ext_dof, 3*self.size), float)
        # translation
        result[0, 0::3] = 1
        result[1, 1::3] = 1
        result[2, 2::3] = 1
        if not self.periodic:
            # rotation
            counter = 3
            for iax in iaxs.transpose():
                result[counter,0::3] = (iax[1]*(self.coordinates[:,2] - self.com[2])
                                       -iax[2]*(self.coordinates[:,1] - self.com[1])) # x
                result[counter,1::3] = (iax[2]*(self.coordinates[:,0] - self.com[0])
                                       -iax[0]*(self.coordinates[:,2] - self.com[2])) # y
                result[counter,2::3] = (iax[0]*(self.coordinates[:,1] - self.com[1])
                                       -iax[1]*(self.coordinates[:,0] - self.com[0])) # z
                counter += 1
        return result

    @cached
    def external_basis(self):
        """The basis for small displacements in the external degrees of freedom.

           The basis is expressed in mass-weighted Cartesian coordinates.
           The three translations are along the x, y, and z-axis. The three
           rotations are about an axis parallel to the x, y, and z-axis
           through the center of mass.
           The returned result depends on the periodicity of the system:

           * When the system is periodic, only the translation external degrees
             are included. The result is an array with shape (3,3N)
           * When the system is not periodic, the rotational external
             degrees are also included. The result is an array with shape (6,3N). The
             first three rows correspond to translation, the latter three rows
             correspond to rotation.
        """
        center = (self.coordinates*self.masses3.reshape((-1,3))).sum(0)/self.mass  # center of mass
        result = transrot_basis(self.coordinates - center, not self.periodic)
        result *= np.sqrt(self.masses3) # transform basis to mass weighted coordinates

        return result

    @cached
    def masses3(self):
        """An array with the diagonal of the mass matrix in Cartesian coordinates.

           Each atom mass is repeated three times. The total length of the
           array is 3N.
        """
        return np.array([self.masses, self.masses, self.masses]).transpose().ravel()

    def get_submolecule(self, selected, energy=None, multiplicity=None, symmetry_number=None, periodic=None, graph=None, title=None, symbols=None, unit_cell=None):
        """Create a submolecule with a selection of atoms

           Argument:
            | ``selected`` -- list of atom indices, numbering starts at 0

           Optional arguments:
            | ``energy`` -- Molecular electronic energy
            | ``multiplicity`` -- The spin multiplicity of the electronic system
            | ``symmetry_number`` -- The rotational symmetry number. Inherited,
                                     None when not known or computed [default=0]
            | ``periodic`` -- True when the system is periodic in three dimensions
                              [default=False]
            | ``title`` -- The title of the system
            | ``graph`` -- The molecular graph of the system
            | ``symbols`` -- A list with atom symbols
            | ``unit_cell`` -- The unit cell vectors for periodic structures

           The function returns a Molecule object consisting of the
           atoms in the selected list. The numbers, coordinates, masses, gradient,
           and Hessian are reduced in size.
           The energy, multiplicity, symmetry_number, periodic and symbols
           attributes are copies of the original molecule attributes, except if
           they are explicitly specified as optional arguments, then they are
           overwritten.
        """
        selected = np.array(selected)
        selected3 = np.array( sum( [[3*at, 3*at+1, 3*at+2] for at in selected] ,[]) )

        # if the following are none, then use the attributes of the original molecule
        if energy is None: energy = self.energy
        if multiplicity is None: multiplicity = self.multiplicity
        if symmetry_number is None: symmetry_number = self.symmetry_number
        if periodic is None: periodic = self.periodic
        if symbols is None and self.symbols is not None:
            symbols = [self.symbols[at] for at in selected]
        if unit_cell is None:
            unit_cell = self.unit_cell

        return Molecule(
            self.numbers[selected],
            self.coordinates[selected,:],
            self.masses[selected],
            energy,
            self.gradient[selected,:],
            self.hessian[selected3,:][:,selected3],
            multiplicity,
            symmetry_number = symmetry_number,
            periodic = periodic,
            title = title,
            graph = graph,
            symbols = symbols,
            unit_cell = unit_cell,
        )

    def write_to_file(self, filename):
        """Write the molecule to a human-readable checkpoint file.

           Argument:
            | ``filename`` -- the file to write to
        """
        from tamkin.io.internal import dump_chk
        data = {}
        for key in "numbers", "coordinates", "masses", "energy", "gradient", \
                   "hessian", "multiplicity", "symmetry_number", "periodic", \
                   "title", "symbols":
            value = getattr(self, key, None)
            if value is not None:
                data[key] = value
        if self.graph is not None:
            data["edges"] = np.array([tuple(edge) for edge in self.graph.edges])
        if self.unit_cell is not None:
            data["cell_vectors"] = self.unit_cell.matrix
            data["cell_active"] = self.unit_cell.active
        dump_chk(filename, data)

    @classmethod
    def read_from_file(cls, filename):
        """Construct a Molecule object from a previously saved checkpoint file

           Arguments:
            | ``filename`` -- the file to load from

           Usage::

             >>> mol = Molecule.read_from_file("mol.chk")

        """
        from tamkin.io.internal import load_chk
        # load the file
        data = load_chk(filename)
        # check the names of the fields:
        mandatory_fields = set([
            "numbers", "coordinates", "masses", "energy", "gradient", "hessian"
        ])
        if not set(data.iterkeys()).issuperset(mandatory_fields):
            raise IOError("The Checkpoint file does not contain the mandatory fields.")
        # take the mandatory fields
        constructor_args = {}
        for mfield in mandatory_fields:
            constructor_args[mfield] = data[mfield]
        # take the optional arguments if present
        opt_fields = ["multiplicity", "symmetry_number", "periodic", "title", "symbols"]
        for ofield in opt_fields:
            if ofield in data:
                constructor_args[ofield] = data[ofield]
        # take the special optional arguments that need conversion
        if "edges" in data:
            graph = MolecularGraph(data["edges"], data["numbers"])
            constructor_args["graph"] = graph
        if "cell_vectors" in data:
            unit_cell = UnitCell(data["cell_vectors"], data.get("cell_active"))
            constructor_args["unit_cell"] = unit_cell
        # construct the molecule object
        return Molecule(**constructor_args)

    def raise_ext(self, shift=1e0):
        """Raise the eigenvalues of the global translations and rotations
        to a high value, such that their coupling with the internal vibrations
        becomes negligible, and they can easily be isolated from the vibrations."""

        # Construct basis for global translations and rotations
        D = self.external_basis.transpose() # mass-weighted
        # make it orthonormal
        svd_threshold = 1e-5
        U, W, Vt = np.linalg.svd(D, full_matrices=False)
        rank = (abs(W) > abs(W[0])*svd_threshold).sum()
        D = U[:,:rank]
        proj = np.dot(D,D.transpose())

        proj1 = proj * (self.masses3.reshape((1,-1)))**(0.5)
        proj2 = proj1* (self.masses3.reshape((-1,1)))**(0.5)

        # Add the shift to the Hessian. The gradient is not changed I guess TODO check this.
        hessian = self.hessian + shift*proj2

        # Use the attributes of the original molecule if they exist
        if hasattr(self,"title"): # check if attribute exists
            title = self.title
        else: title = None
        if hasattr(self,"graph"): # check if attribute exists
            graph = self.graph
        else: graph = None
        if hasattr(self,"symbols"): # check if attribute exists
            symbols = self.symbols
        else: symbols = None
        if hasattr(self,"unit_cell"): # check if attribute exists
            unit_cell = self.unit_cell
        else: unit_cell = None

        return Molecule(
            self.numbers,
            self.coordinates,
            self.masses,
            self.energy,
            self.gradient,
            hessian,             # replace Hessian by new one
            self.multiplicity,
            symmetry_number = self.symmetry_number,
            periodic = self.periodic,
            title = title,
            graph = graph,
            symbols = symbols,
            unit_cell = unit_cell,
        )

    def constrain_ext(self):
        """Project the global translational and rotational vectors
        out of the Hessian and the gradient and return a new Molecule instance."""

        # Construct projector
        D = self.external_basis.transpose() # mass-weighted
        # make it orthonormal
        svd_threshold = 1e-5
        U, W, Vt = np.linalg.svd(D, full_matrices=False)
        rank = (abs(W) > abs(W[0])*svd_threshold).sum()
        D = U[:,:rank]
        proj = np.dot(D,D.transpose())
        #print np.sum((proj-np.dot(proj,proj))**2)

        proj1 = proj * (self.masses3.reshape((1,-1)))**(0.5)
        proj2 = proj1* (self.masses3.reshape((-1,1)))**(-0.5)
        projL = np.identity(len(D)) - proj2.transpose()
        projR = np.identity(len(D)) - proj2
        # test
        #print np.sum((projR-np.dot(projR,projR))**2)
        #print np.sum((projL-np.dot(projL,projL))**2)

        # Project hessian and gradient
        hessian = np.dot(np.dot(projL,self.hessian),projR)
        gradient = np.dot(projL, self.gradient.reshape((-1,1))).reshape((-1,3))

        # Use the attributes of the original molecule if they exist
        if hasattr(self,"title"): # check if attribute exists
            title = self.title
        else: title = None
        if hasattr(self,"graph"): # check if attribute exists
            graph = self.graph
        else: graph = None
        if hasattr(self,"symbols"): # check if attribute exists
            symbols = self.symbols
        else: symbols = None
        if hasattr(self,"unit_cell"): # check if attribute exists
            unit_cell = self.unit_cell
        else: unit_cell = None

        return Molecule(
            self.numbers,
            self.coordinates,
            self.masses,
            self.energy,
            gradient,       # replace gradient by new one
            hessian,             # replace Hessian by new one
            self.multiplicity,
            symmetry_number = self.symmetry_number,
            periodic = self.periodic,
            title = title,
            graph = graph,
            symbols = symbols,
            unit_cell = unit_cell,
        )


class BareNucleus(Molecule):
    """A molecule object for bare nuclei"""

    def __init__(self, number, mass=None):
        """
           Argument:
            | ``number`` -- The atom number

           Optional argument:
            | ``mass`` -- The mass of the atom in atomic units
        """
        if mass is None:
            mass = periodic[number].mass
        Molecule.__init__(
            self, [number], [[0,0,0]], [mass], 0.0, [[0,0,0]],
            [[0,0,0],[0,0,0],[0,0,0]], 1, 0, False
        )


class Proton(BareNucleus):
    """A molecule object for a proton (or one of its isotopes)"""

    def __init__(self, mass=None):
        """
           Optional argument:
            | ``mass`` -- The mass of the proton in atomic units
        """
        BareNucleus.__init__(self, 1, mass)


class RotScan(ReadOnly):
    """A container for rotational scan data"""

    def check_top_indexes(self, top_indexes):
        if len(top_indexes) < 1:
            raise TypeError("A rotational scan must have at least one atom "
                "rotating")

    dihedral = ReadOnlyAttribute(np.ndarray, none=False, npdim=1,
        npshape=(4,), npdtype=int)
    top_indexes = ReadOnlyAttribute(np.ndarray, none=False,
        check=check_top_indexes, npdim=1, npdtype=int)
    potential = ReadOnlyAttribute(np.ndarray, npdim=2, npshape=(2, None),
        npdtype=float)

    def __init__(self, dihedral, molecule=None, top_indexes=None, potential=None):
        """
           Arguments
            | ``dihedral`` -- the index of the atoms that define the dihedral
                              angle

           Optional arguments
            | ``molecule`` -- a molecule object. required when top_indexes is not
                              given
            | ``top_indexes`` -- a list of atom indexes involved in the rotor.
                                 required when molecule is not given or when the
                                 top_indexes can not be derived automatically.
            | ``potential`` -- rotational potential info (if this is a hindered
                               rotor). must be a two-tuple containing the angles
                               and the corresponding energies.

        """
        self.dihedral = dihedral
        if top_indexes is None:
            # try to deduce the top indexes
            if molecule is None:
                raise ValueError("missing arguemt: top_indexes or molecule from which top_indexes can be derived")
            atom0, atom1 = dihedral[1:3]
            graph = MolecularGraph.from_geometry(molecule)
            half0, half1 = graph.get_halfs(atom0, atom1)
            if (len(half0) + len(half1) != molecule.size) or len(half0) == 1 or len(half1) == 1:
                raise ValueError("The rotating top could not be assigned properly. Specify the top_indexes manually.")
            if len(half1) > len(half0):
                top_indexes = half0
                top_indexes.discard(atom0)
            else:
                top_indexes = half1
                top_indexes.discard(atom1)
            top_indexes = tuple(top_indexes)
        else:
            # check the sanity of the top indexes
            if not ((self.dihedral[0] in top_indexes) ^ (self.dihedral[3] in top_indexes)):
                raise ValueError('The top must contain either first or the last atom of the dihedral angle.')
            if ((self.dihedral[1] in top_indexes) | (self.dihedral[2] in top_indexes)):
                raise ValueError('The top may not contain atoms from the central bond of the dihedral angle.')
        self.top_indexes = top_indexes
        self.potential = potential


def translate_pbc(molecule, selected, displ, vectors = None):
    """Translate the structure along the lattice vectors

       This method is meant to be used in periodic structures, where periodic
       boundary conditions apply (pbc).

       Arguments:
        | ``molecule`` -- A Molecule instance.
        | ``selected`` -- A list of indices of the atoms that will be displaced.
        | ``displ`` -- A list of 3 integers: [i0,i1,i2]. The selected atoms will
                       be displaced over i0 lattice distances in the 0-axis
                       direction, similarly for i1 and i2.

       Optional argument:
        | ``vectors`` -- The lattice vectors, one in each column. If not
                         specified, the vectors in the unit_cell attribute of
                         the molecule is used.
    """
    coordinates = molecule.coordinates.copy()
    if vectors is None:
        if molecule.unit_cell is None:
            raise ValueError("No vectors for axes known")
        else:
            vectors = molecule.unit_cell.matrix

    # translate
    for atom in selected:
        for axis in range(3):
            coordinates[atom,:] += displ[axis]*vectors[:,axis]

    return molecule.copy_with(coordinates=coordinates)



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


import numpy as np
import os
import pkg_resources
import unittest

from molmod.periodic import periodic
from molmod.units import angstrom, amu, calorie, avogadro, electronvolt
from molmod.constants import lightspeed
from molmod.test.common import tmpdir

from tamkin import *


__all__ = ["IOTestCase"]


class IOTestCase(unittest.TestCase):
    def test_load_fixed_g03com(self):
        fixed_atoms = load_fixed_g03com(
            pkg_resources.resource_filename(__name__, "../data/test/mat/Zp_p_prod.18aug.com"))
        self.assertEqual(len(fixed_atoms), 48)
        self.assertEqual(fixed_atoms, range(114,114+48))

    def test_load_molecule_g03fchk(self):
        atoms = 181
        molecule = load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/mat/Zp_p_react.28aug.fchk"))
        self.assertEqual(molecule.hessian.shape,(atoms*3,atoms*3))
        self.assertAlmostEqual(molecule.energy, -3053.805846445570, 7)
        molecule = load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/mat/Zp_p_react.28aug.fchk"),
            energy=-123.0)
        self.assertAlmostEqual(molecule.energy, -123.0, 7)
        molecule = load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/mat/Zp_p_react.28aug.fchk"),
            pkg_resources.resource_filename(__name__, "../data/test/mat/Zp_p_react.14mei.fchk"))
        self.assertAlmostEqual(molecule.energy, -18613.135744186180, 7)

    def test_load_molecule_g98fchk(self):
        atoms = 6
        molecule = load_molecule_g98fchk(
            pkg_resources.resource_filename(__name__, "../data/test/g98/freqs.fchk"))
        self.assertEqual(molecule.hessian.shape,(atoms*3,atoms*3))
        self.assertAlmostEqual(molecule.masses[0]/amu, 12.011)
        self.assertAlmostEqual(molecule.masses[2]/amu, 1.0079)
        self.assertAlmostEqual(molecule.energy, -78.58745828877478, 7)
        molecule = load_molecule_g98fchk(
            pkg_resources.resource_filename(__name__, "../data/test/g98/freqs.fchk"),
            energy=-123.0)
        self.assertAlmostEqual(molecule.energy, -123.0, 7)

    def test_load_molecule_g03fchkvdw(self):
        atoms = 179
        molecule = load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/matvdw/R.fchk"),
            pkg_resources.resource_filename(__name__, "../data/test/matvdw/R_SCF.fchk"),
            pkg_resources.resource_filename(__name__, "../data/test/matvdw/R_b3lyp-d.out"))

        self.assertEqual(molecule.hessian.shape,(atoms*3,atoms*3))
        self.assertAlmostEqual(molecule.energy,-18612.352569964281 , 7)

    def test_load_fixed_fchk(self):
        molecule = load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/ethane/gaussian.fchk"))
        assert (molecule.fixed == [0, 6, 7]).all()

    def test_load_molecule_cp2k(self):
        molecule = load_molecule_cp2k(
            pkg_resources.resource_filename(__name__, "../data/test/cp2k/pentane/sp.out"),
            pkg_resources.resource_filename(__name__, "../data/test/cp2k/pentane/freq.out"))
        self.assertAlmostEqual(molecule.energy, 0.012255059530862)
        self.assertEqual(molecule.multiplicity, 1)
        self.assertEqual(molecule.numbers[0], 6)
        self.assertEqual(molecule.numbers[4], 1)
        self.assertAlmostEqual(molecule.masses[0], periodic[6].mass, 5)
        self.assertAlmostEqual(molecule.masses[4], periodic[1].mass, 0)
        self.assertAlmostEqual(molecule.coordinates[5,1]/angstrom, 13.928520)
        self.assertAlmostEqual(molecule.gradient[0,2], 0.0000000038, 9)
        self.assertAlmostEqual(molecule.gradient[11,0], 0.0000000177, 9)
        self.assertAlmostEqual(molecule.hessian[0,0], 49.629809*1e-6*molecule.masses[0], 6)
        self.assertAlmostEqual(molecule.hessian[-1,-1], 287.884198*1e-6*molecule.masses[-1], 6)
        self.assertAlmostEqual(molecule.unit_cell.matrix[0,0]/angstrom, 30.000,3)
        self.assertAlmostEqual(molecule.unit_cell.matrix[1,2]/angstrom, 0.000,3)

    def test_load_molecule_cp2k_23(self):
        molecule = load_molecule_cp2k(
            pkg_resources.resource_filename(__name__, "../data/test/cp2k/john/scf.out"),
            pkg_resources.resource_filename(__name__, "../data/test/cp2k/john/quickie.out"))
        self.assertAlmostEqual(molecule.energy, -141.189006820072791)
        self.assertEqual(molecule.multiplicity, 1)
        self.assertEqual(molecule.numbers[0], 14)
        self.assertEqual(molecule.numbers[4], 8)
        self.assertAlmostEqual(molecule.masses[0], periodic[14].mass)
        self.assertAlmostEqual(molecule.masses[4], periodic[8].mass)
        self.assertAlmostEqual(molecule.coordinates[5,1]/angstrom, 12.150867)
        self.assertAlmostEqual(molecule.gradient[0,2], -0.00008196, 9)
        self.assertAlmostEqual(molecule.gradient[11,0], -0.00041872, 9)
        self.assertAlmostEqual(molecule.hessian[0,0], 8.297378*1e-6*molecule.masses[0], 6)
        self.assertAlmostEqual(molecule.hessian[-1,-1], 71.133554*1e-6*molecule.masses[-1], 6)
        self.assertAlmostEqual(molecule.unit_cell.matrix[0,0]/angstrom, 20.000,3)
        self.assertAlmostEqual(molecule.unit_cell.matrix[1,2]/angstrom, 0.000,3)

    def test_load_molecule_cp2k_dan_phva(self):
        molecule = load_molecule_cp2k(
            pkg_resources.resource_filename(__name__, "../data/test/cp2k/dan/cp2k.out"),
            pkg_resources.resource_filename(__name__, "../data/test/cp2k/dan/freq.out"))
        self.assertAlmostEqual(molecule.energy, -290.409097595743333)
        self.assertEqual(molecule.multiplicity, 1)
        self.assertEqual(molecule.numbers[0], 6)
        self.assertEqual(molecule.numbers[4], 6)
        self.assertEqual(molecule.numbers[-1], 1)
        self.assertAlmostEqual(molecule.masses[0], periodic[6].mass, 0)
        self.assertAlmostEqual(molecule.masses[-1], periodic[1].mass, 0)
        self.assertAlmostEqual(molecule.coordinates[5,1], -2.458266*angstrom)
        self.assertAlmostEqual(molecule.gradient[-1,0], 0.00000044, 9)
        self.assertAlmostEqual(molecule.gradient[-3,2], 0.00000080, 9)
        self.assertAlmostEqual(molecule.hessian[0,0], 0.0, 6)
        self.assertAlmostEqual(molecule.hessian[0,-1], 0.0, 6)
        self.assertAlmostEqual(molecule.hessian[-1,0], 0.0, 6)
        self.assertAlmostEqual(molecule.hessian[-18,-1],
            0.5*(-0.418336-0.418196)*1e-6*(molecule.masses[-1]*molecule.masses[-18])**0.5, 6)
        self.assertAlmostEqual(molecule.hessian[-18,-18], 19.137757*1e-6*molecule.masses[-18], 6)
        self.assertAlmostEqual(molecule.hessian[-1,-1], 51.483068*1e-6*molecule.masses[-1], 6)
        self.assertAlmostEqual(molecule.unit_cell.matrix[0,0], 10.657*angstrom, 3)
        self.assertAlmostEqual(molecule.unit_cell.matrix[1,0], -6.153*angstrom, 3)

    def test_load_fixed_cp2k(self):
        fixed = load_fixed_cp2k(
            pkg_resources.resource_filename(__name__, "../data/test/cp2k/dan/freq.out"))
        np.testing.assert_equal(fixed, np.arange(52 - 6))

    def test_load_molecule_cpmd(self):
        molecule = load_molecule_cpmd(
            pkg_resources.resource_filename(__name__, "../data/test/cpmd/damp.out"),
            pkg_resources.resource_filename(__name__, "../data/test/cpmd/GEOMETRY.xyz"),
            pkg_resources.resource_filename(__name__, "../data/test/cpmd/MOLVIB"))
        self.assertAlmostEqual(molecule.energy, -17.14142079)
        self.assertEqual(molecule.multiplicity, 1)
        self.assertEqual(molecule.numbers[0], 8)
        self.assertEqual(molecule.numbers[2], 1)
        self.assertAlmostEqual(molecule.masses[0]/amu, 15.999400)
        self.assertAlmostEqual(molecule.masses[2]/amu, 1.007970)
        self.assertAlmostEqual(molecule.coordinates[2,0]/angstrom, .907207019799)
        self.assertAlmostEqual(molecule.gradient[0,2], 0.0)
        self.assertAlmostEqual(molecule.gradient[2,0], 0.0)
        self.assertAlmostEqual(molecule.hessian[0,0], 0.530679165332463953, 6)
        self.assertAlmostEqual(molecule.hessian[-1,-1], 0.921045226686428159E-01, 6)

    def test_load_molecule_charmm(self):
        molecule = load_molecule_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.cor"),
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.hess.full"))
        self.assertAlmostEqual(molecule.energy/(1000*calorie/avogadro), -2.1303308955)
        self.assertEqual(molecule.multiplicity, 1)
        self.assertEqual(molecule.numbers[0], 6)
        self.assertEqual(molecule.numbers[4], 1)
        self.assertAlmostEqual(molecule.masses[0]/amu, 12.01100)
        self.assertAlmostEqual(molecule.masses[4]/amu, 1.00800)
        self.assertAlmostEqual(molecule.coordinates[5,1]/angstrom, 1.3582528196)
        self.assertAlmostEqual(molecule.gradient[0,2]/(1000*calorie/avogadro/angstrom), -0.0000000007, 9)
        self.assertAlmostEqual(molecule.gradient[8,0]/(1000*calorie/avogadro/angstrom), -0.0000001462, 9)
        self.assertAlmostEqual(molecule.hessian[0,0]/(1000*calorie/avogadro /angstrom**2), 1409.7091337384, 6)
        self.assertAlmostEqual(molecule.hessian[-1,-1]/(1000*calorie/avogadro /angstrom**2), 474.7950312957, 6)

    def test_load_molecule_qchem(self):
        molecule = load_molecule_qchem(
            pkg_resources.resource_filename(__name__, "../data/test/qchem/h2o2.hf.sto-3g.freq.out"),
            hessfile=pkg_resources.resource_filename(__name__, "../data/test/qchem/hessian.dat"))
        self.assertAlmostEqual(molecule.energy, -148.7649966058)
        self.assertEqual(molecule.multiplicity, 1)
        self.assertEqual(molecule.numbers[0], 1)
        self.assertEqual(molecule.numbers[3], 8)
        self.assertAlmostEqual(molecule.masses[0]/amu,  1.00783, 5)
        self.assertAlmostEqual(molecule.masses[3]/amu, 15.99491, 5)
        self.assertAlmostEqual(molecule.coordinates[2,1]/angstrom, -0.688720)
        self.assertAlmostEqual(molecule.gradient[0,2]/(1000*calorie/avogadro/angstrom), -0.00, 5)
        self.assertAlmostEqual(molecule.gradient[3,0]/(1000*calorie/avogadro/angstrom), -0.00, 5)
        self.assertAlmostEqual(molecule.hessian[0,0]/(1000*calorie/avogadro/angstrom**2), 364.769480916757800060, 6)
        self.assertAlmostEqual(molecule.hessian[-1,-1]/(1000*calorie/avogadro/angstrom**2), 338.870127396983150447, 6)

    def test_load_molecule_vasp_53(self):
        molecule = load_molecule_vasp(
            pkg_resources.resource_filename(__name__, "../data/test/lucas/vasp_5_3_5_complex/CONTCAR_opt"),
            pkg_resources.resource_filename(__name__, "../data/test/lucas/vasp_5_3_5_complex/OUTCAR_freq"))
        # contcar
        assert molecule.numbers[0] == 6
        assert (molecule.numbers[1:] == 1).all()
        assert molecule.size == 5
        assert molecule.unit_cell.matrix[0,0] == 15.0*angstrom
        assert molecule.unit_cell.matrix[1,2] == 0.0
        self.assertAlmostEqual(molecule.coordinates[0,0]/angstrom, 7.15840, 3)
        self.assertAlmostEqual(molecule.coordinates[1,2]/angstrom, 8.44640, 2) #?
        self.assertAlmostEqual(molecule.coordinates[-1,-1]/angstrom, 6.95131, 2) #?
        # outcar_freq
        assert molecule.masses[0] == 12.011*amu
        assert (molecule.masses[1:] == 1.000*amu).all()
        hunit = electronvolt/angstrom**2
        assert molecule.hessian[0,0] == 53.624756*hunit
        assert molecule.hessian[-1,-1] == 31.299419*hunit
        self.assertAlmostEqual(molecule.hessian[2,5], 0.5*(-7.551817 + 3.319877)*hunit)
        assert molecule.energy == -24.11901936*electronvolt
        gunit = electronvolt/angstrom
        assert molecule.gradient[0, 0] == 0.096977*gunit
        assert molecule.gradient[2, 1] == 0.100275*gunit
        assert molecule.gradient[-1, -1] == -0.212810*gunit

    def test_load_molecule_vasp_5_3_5_gamma(self):
        molecule = load_molecule_vasp(
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_3_5_gamma/CONTCAR_opt"),
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_3_5_gamma/OUTCAR_freq"))
        # contcar
        assert molecule.numbers[0] == 6
        assert (molecule.numbers[1:] == 1).all()
        assert molecule.size == 5
        assert molecule.unit_cell.matrix[0,0] == 15.0*angstrom
        assert molecule.unit_cell.matrix[1,2] == 0.0
        self.assertAlmostEqual(molecule.coordinates[0,0]/angstrom, 7.15782, 3)
        self.assertAlmostEqual(molecule.coordinates[1,2]/angstrom, 8.44278, 1) #??
        self.assertAlmostEqual(molecule.coordinates[-1,-1]/angstrom, 6.95393, 2)  #?
        # outcar_freq
        assert molecule.masses[0] == 12.011*amu
        assert (molecule.masses[1:] == 1.000*amu).all()
        hunit = electronvolt/angstrom**2
        assert molecule.hessian[0,0] == 47.756815*hunit
        assert molecule.hessian[-1,-1] == 31.561376*hunit
        self.assertAlmostEqual(molecule.hessian[2,5], 0.5*(-2.265871 + -3.645039)*hunit)
        assert molecule.energy == -24.12364199*electronvolt
        gunit = electronvolt/angstrom
        assert molecule.gradient[0, 0] == -0.005459*gunit
        assert molecule.gradient[2, 1] == -0.008215*gunit
        assert molecule.gradient[-1, -1] == 0.003424*gunit

    def test_load_molecule_vasp_5_3_5_gamma_part(self):
        molecule = load_molecule_vasp(
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_3_5_gamma/CONTCAR_opt"),
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_3_5_gamma/OUTCAR_freq_part"))
        # outcar_freq
        hunit = electronvolt/angstrom**2
        assert molecule.hessian[0,0] == 0.0
        assert molecule.hessian[2,5] == 0.0
        assert molecule.hessian[-1,-1] == 31.561374*hunit
        self.assertAlmostEqual(molecule.hessian[6,9], 0.5*(-2.601094 + -2.794160)*hunit)

    def test_load_molecule_vasp_5_2_11_complex(self):
        molecule = load_molecule_vasp(
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_2_11_complex/CONTCAR_opt"),
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_2_11_complex/OUTCAR_freq"))
        # outcar_freq
        assert molecule.masses[0] == 12.011*amu
        assert (molecule.masses[1:] == 1.000*amu).all()
        hunit = electronvolt/angstrom**2
        assert molecule.hessian[0,0] == 47.762604*hunit
        assert molecule.hessian[-1,-1] == 31.565279*hunit
        self.assertAlmostEqual(molecule.hessian[2,5], 0.5*(-3.648356 + -2.264335)*hunit)
        assert molecule.energy == -24.123642*electronvolt
        gunit = electronvolt/angstrom
        assert molecule.gradient[0, 0] == -0.005432*gunit
        assert molecule.gradient[2, 1] == -0.008204*gunit
        assert molecule.gradient[-1, -1] == 0.003425*gunit

    def test_load_molecule_vasp_5_2_11_complex_part(self):
        molecule = load_molecule_vasp(
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_2_11_complex/CONTCAR_opt"),
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_2_11_complex/OUTCAR_freq_part"))
        # outcar_freq
        hunit = electronvolt/angstrom**2
        assert molecule.hessian[0,0] == 0.0
        assert molecule.hessian[2,5] == 0.0
        assert molecule.hessian[-1,-1] == 31.565272*hunit
        self.assertAlmostEqual(molecule.hessian[6,9], 0.5*(-2.600373 + -2.836454)*hunit)
        assert molecule.energy == -24.123642*electronvolt

    def test_load_molecule_vasp_5_3_3_complex(self):
        molecule = load_molecule_vasp(
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_3_3_complex/CONTCAR_opt"),
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_3_3_complex/OUTCAR_freq"))
        # outcar_freq
        assert molecule.masses[0] == 12.011*amu
        assert (molecule.masses[1:] == 1.000*amu).all()
        hunit = electronvolt/angstrom**2
        assert molecule.hessian[0,0] == 47.757096*hunit
        assert molecule.hessian[-1,-1] == 31.561341*hunit
        self.assertAlmostEqual(molecule.hessian[2,5], 0.5*(-3.645047 + -2.265767)*hunit)
        assert molecule.energy == -24.12364179*electronvolt
        gunit = electronvolt/angstrom
        assert molecule.gradient[0, 0] == -0.005431*gunit
        assert molecule.gradient[2, 1] == -0.008279*gunit
        assert molecule.gradient[-1, -1] == 0.003335*gunit

    def test_load_molecule_vasp_5_3_3_complex_part(self):
        molecule = load_molecule_vasp(
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_3_3_complex/CONTCAR_opt"),
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_3_3_complex/OUTCAR_freq_part"))
        # outcar_freq
        hunit = electronvolt/angstrom**2
        assert molecule.hessian[0,0] == 0.0
        assert molecule.hessian[2,5] == 0.0
        assert molecule.hessian[-1,-1] == 31.561353*hunit
        self.assertAlmostEqual(molecule.hessian[6,9], 0.5*(-2.601060 + -2.794022)*hunit)

    def test_load_molecule_vasp_5_3_5_gamma_part_energy(self):
        molecule = load_molecule_vasp(
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_3_5_gamma/CONTCAR_opt"),
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_3_5_gamma/OUTCAR_freq_part"),
            energy=1.476)
        assert molecule.energy == 1.476

    def test_load_molecule_vasp_5_3_5_gamma_part_outcar_energy(self):
        molecule = load_molecule_vasp(
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_3_5_gamma/CONTCAR_opt"),
            pkg_resources.resource_filename(__name__, "../data/test/julianna/vasp_5_3_5_gamma/OUTCAR_freq_part"),
            pkg_resources.resource_filename(__name__, "../data/test/lucas/vasp_5_3_5_complex/OUTCAR_freq"))
        assert molecule.energy == -24.11901936*electronvolt

    def test_checkpoint(self):
        molecule = load_molecule_cp2k(
            pkg_resources.resource_filename(__name__, "../data/test/cp2k/pentane/sp.out"),
            pkg_resources.resource_filename(__name__, "../data/test/cp2k/pentane/freq.out"))
        nma1 = NMA(molecule)
        with tmpdir(__name__, 'test_checkpoint') as dn:
            fn_out = os.path.join(dn, 'test.chk')
            nma1.write_to_file(fn_out)
            nma2 = NMA.read_from_file(fn_out)

        self.assertEqual(nma1.freqs.shape, nma2.freqs.shape)
        self.assertEqual(nma1.modes.shape, nma2.modes.shape)
        self.assertEqual(nma1.masses.shape, nma2.masses.shape)
        self.assertEqual(nma1.numbers.shape, nma2.numbers.shape)
        self.assertEqual(nma1.coordinates.shape, nma2.coordinates.shape)
        self.assertEqual(nma1.inertia_tensor.shape, nma2.inertia_tensor.shape)

        self.assert_(abs(nma1.freqs - nma2.freqs).max()/abs(nma1.freqs).max() < 1e-15)
        self.assert_(abs(nma1.modes - nma2.modes).max()/abs(nma1.modes).max() < 1e-15)
        self.assert_(abs(nma1.masses - nma2.masses).max()/abs(nma1.masses).max() < 1e-15)
        self.assert_(abs(nma1.coordinates - nma2.coordinates).max()/abs(nma1.coordinates).max() < 1e-15)
        self.assert_(abs(nma1.inertia_tensor - nma2.inertia_tensor).max()/abs(nma1.inertia_tensor).max() < 1e-15)
        self.assert_((nma1.numbers==nma2.numbers).all())

        self.assertAlmostEqual(nma1.mass, nma2.mass)
        self.assertAlmostEqual(nma1.energy, nma2.energy)
        self.assertEqual(nma1.multiplicity, nma2.multiplicity)
        self.assertEqual(nma1.symmetry_number, nma2.symmetry_number)

    def test_load_indices(self):
        blocks = load_indices(pkg_resources.resource_filename(__name__, "../data/test/an/fixed.07.txt"), groups=True)
        self.assertEqual(blocks, [[3,2,6]])
        blocks = load_indices(pkg_resources.resource_filename(__name__, "../data/test/an/fixed.07.txt"))
        self.assertEqual(blocks, [3,2,6])
        blocks = load_indices(pkg_resources.resource_filename(__name__, "../data/test/an/fixed.08.txt"))
        self.assertEqual(blocks, [5,4,8])

    def test_load_indices_ranges(self):
        blocks = load_indices(pkg_resources.resource_filename(__name__, "../data/test/an/fixed_ranges.txt"), groups=True)
        self.assertEqual(blocks, [[0, 2, 3, 4], [9, 10, 11, 12, 13, 19], [21]])
        blocks = load_indices(pkg_resources.resource_filename(__name__, "../data/test/an/fixed_ranges.txt"), shift=0)
        self.assertEqual(blocks, [1, 3, 4, 5, 10, 11, 12, 13, 14, 20, 22])
        blocks = load_indices(pkg_resources.resource_filename(__name__, "../data/test/an/fixed_ranges_shift.txt"), groups=True)
        self.assertEqual(blocks, [[1, 3, 4, 5], [10, 11, 12, 13, 14, 20], [22]])
        blocks = load_indices(pkg_resources.resource_filename(__name__, "../data/test/an/fixed_ranges_shift.txt"), shift=0)
        self.assertEqual(blocks, [1, 3, 4, 5, 10, 11, 12, 13, 14, 20, 22])

    def test_dump_modes_xyz(self):
        molecule = load_molecule_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.cor"),
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.hess.full"))
        nma = NMA(molecule)
        with tmpdir(__name__, 'test_dump_modes_xyz') as dn:
            prefix = os.path.join(dn, 'mode')
            dump_modes_xyz(nma, 6, prefix=prefix, amplitude=50.0)
            with open(prefix + ".6.xyz") as f:
                # 1st line
                line = f.readline().strip()
                self.assertEqual(line,"9")
                # 2nd line
                line = f.readline().strip()
                self.assertEqual(line,"frame 0")
                # 3rd line
                line = f.readline()
                words = line.split()
                self.assertEqual(len(words),4)
                self.assertEqual(words[0],"C")
                self.assertEqual(words[2],"0.081608346")
                for i in range(9):
                    line = f.readline().strip()
                self.assertEqual(line,"9")

    def test_dump_modes_gaussian(self):
        molecule = load_molecule_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.cor"),
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.hess.full"))
        nma = NMA(molecule)
        with tmpdir(__name__, 'test_dump_modes_gaussian') as dn:
            # blind test, no double checking of the output file
            fn_log = os.path.join(dn, 'modes_gaussian.log')
            dump_modes_gaussian(fn_log, nma)

    def test_load_dump_indices1(self):
        subs = range(10)
        with tmpdir(__name__, 'test_load_dump_indices1') as dn:
            dump_indices(os.path.join(dn, "subs-atoms.1.txt"), subs, shift=0)
            dump_indices(os.path.join(dn, "subs-atoms.2.txt"), subs, shift=1)
            dump_indices(os.path.join(dn, "subs-atoms.3.txt"), subs)

            subs1 = load_indices(os.path.join(dn, "subs-atoms.1.txt"), shift=0)
            subs2 = load_indices(os.path.join(dn, "subs-atoms.2.txt"), shift=-1)
            subs22 = load_indices(os.path.join(dn, "subs-atoms.2.txt"))  # should not matter
            subs3 = load_indices(os.path.join(dn, "subs-atoms.3.txt"))

            blocks = [range(10), range(10,20)]
            dump_indices(os.path.join(dn, "blocks.1.txt"), blocks, shift=0)
            dump_indices(os.path.join(dn, "blocks.2.txt"), blocks, shift=1)
            dump_indices(os.path.join(dn, "blocks.3.txt"), blocks)

            blocks1 = load_indices(os.path.join(dn, "blocks.1.txt"), shift=0, groups=True)
            blocks2 = load_indices(os.path.join(dn, "blocks.2.txt"), shift=-1, groups=True)
            blocks22 = load_indices(os.path.join(dn, "blocks.2.txt"), groups=True)  # should not matter
            blocks3 = load_indices(os.path.join(dn, "blocks.3.txt"), groups=True)

        self.assertEqual(len(subs),len(subs1))
        for (i,j) in zip(subs,subs1):
            self.assertEqual(i,j)
        self.assertEqual(len(subs),len(subs2))
        for i,j in zip(subs,subs2):
            self.assertEqual(i,j)
        self.assertEqual(len(subs),len(subs22))
        for i,j in zip(subs,subs22):
            self.assertEqual(i,j)
        self.assertEqual(len(subs),len(subs3))
        for i,j in zip(subs,subs3):
            self.assertEqual(i,j)

        self.assertEqual(len(blocks),len(blocks1))
        for bl,bl1 in zip(blocks,blocks1):
            for i,j in zip(bl,bl1):
                self.assertEqual(i,j)
        self.assertEqual(len(blocks),len(blocks2))
        for bl,bl1 in zip(blocks,blocks2):
            for i,j in zip(bl,bl1):
                self.assertEqual(i,j)
        self.assertEqual(len(blocks),len(blocks22))
        for bl,bl1 in zip(blocks,blocks22):
            for i,j in zip(bl,bl1):
                self.assertEqual(i,j)
        self.assertEqual(len(blocks),len(blocks3))
        for bl,bl1 in zip(blocks,blocks3):
            for i,j in zip(bl,bl1):
                self.assertEqual(i,j)

    def test_load_dump_indices2(self):
        randint = np.random.randint
        for counter in xrange(20):
            for compact in True, False:
                shift = randint(-5,5)
                indices = []
                for i in xrange(randint(10,20)):
                    l = list(set(randint(0,10,randint(20))))
                    if len(l) > 0:
                        indices.append(l)
                indices_flat = sum(indices, [])

                with tmpdir(__name__, '{}_{}'.format('test_load_dump_indices2', counter)) as dn:
                    dump_indices(os.path.join(dn, "indices_blocks.txt"), indices, compact=compact, shift=shift)
                    dump_indices(os.path.join(dn, "indices_flat.txt"), indices_flat, compact=compact, shift=shift)

                    check = load_indices(os.path.join(dn, "indices_blocks.txt"), shift=-shift, groups=True)
                    self.assertEqual(indices, check)
                    check = load_indices(os.path.join(dn, "indices_blocks.txt"), shift=-shift)
                    self.assertEqual(indices_flat, check)
                    check = load_indices(os.path.join(dn, "indices_flat.txt"), shift=-shift, groups=True)
                    self.assertEqual([indices_flat], check)
                    check = load_indices(os.path.join(dn, "indices_flat.txt"), shift=-shift)
                    self.assertEqual(indices_flat, check)

    def test_punch(self):
        mol0 = load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/punch/gaussian.fchk"))
        mol1 = load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/punch/gaussian.fchk"),
            fn_punch=pkg_resources.resource_filename(__name__, "../data/test/punch/fort.7"))
        assert abs(mol0.gradient - mol1.gradient).max() < 1e-8
        assert abs(mol0.hessian - mol1.hessian).max() < 1e-8

    def test_dftd3(self):
        assert load_dftd3(
            pkg_resources.resource_filename(__name__, "../data/test/dftd3/dftd3.out")) == -0.00330057

    def test_dftd_orca(self):
        assert load_dftd_orca(
            pkg_resources.resource_filename(__name__, "../data/test/matvdw/R_b3lyp-d.out")) == -0.404083275



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


# Import the tamkin libarary.
from tamkin import *
from molmod import *

# Load the gaussian data.
molecule = load_molecule_g03fchk("gaussian.fchk")
print molecule.numbers
# Perform the normal mode analysis
nma = NMA(molecule)
# Construct a partition function object with the typical gas phase contributions.
pf = PartFun(nma, [ExtTrans(), ExtRot()])
# 6 is the rotational symmetry number.

# Write some general information about the molecule and the partition function
# to a file.
pf.write_to_file("partfun.txt")
print pf.free_energy(300)/(kjmol)
# Write an extensive overview of the thermodynamic properties to a file:
ta = ThermoAnalysis(pf, [300,400,500,600])
ta.write_to_file("thermo.csv")



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


import os
import pkg_resources
import numpy as np
import unittest

from molmod.units import kjmol, atm, meter, mol, second
from molmod.constants import boltzmann
from molmod.test.common import tmpdir

from tamkin import *


__all__ = ["PFToolsTestCase"]


class PFToolsTestCase(unittest.TestCase):
    def test_reaction_analysis_sterck(self):
        pf_react1 = PartFun(NMA(load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/sterck/aa.fchk"))),
            [ExtTrans(cp=False), ExtRot(1)])
        pf_react2 = PartFun(NMA(load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/sterck/aarad.fchk"))),
            [ExtTrans(cp=False), ExtRot(1)])
        pf_ts = PartFun(NMA(load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/sterck/paats.fchk"))),
            [ExtTrans(cp=False), ExtRot(1)])

        km = KineticModel([pf_react1, pf_react2], pf_ts)
        ra = ReactionAnalysis(km, 280, 360)
        # not a very accurate check because the fit is carried out differently
        # in the fancy excel file where these numbers come from.
        self.assertAlmostEqual(ra.Ea/kjmol, 25.96, 1)
        self.assertAlmostEqual(km.unit, meter**3/mol/second)
        self.assertEqual(km.unit_name, "m^3 mol^-1 s^-1")
        self.assertAlmostEqual(np.log(ra.A/km.unit), np.log(2.29E+02), 1)

        with tmpdir(__name__, 'test_reaction_analysis_sterck') as dn:
            ra.plot_arrhenius(os.path.join(dn, "arrhenius_aa.png"))
            ra.monte_carlo()
            ra.write_to_file(os.path.join(dn, "reaction_aa.txt"))
            ra.plot_parameters(os.path.join(dn, "parameters_aa.png"))

    def test_reaction_analysis_mat(self):
        pf_react = PartFun(NMA(load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/mat5T/react.fchk"))), [])
        pf_ts = PartFun(NMA(load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/mat5T/ts.fchk"))), [])

        km = KineticModel([pf_react], pf_ts)
        ra = ReactionAnalysis(km, 100, 1200, temp_step=50)
        # not a very accurate check because the fit is carried out differently
        # in the fancy excel file where these numbers come from.
        self.assertAlmostEqual(ra.Ea/kjmol, 160.6, 0)
        self.assertAlmostEqual(km.unit, 1.0/second)
        self.assertEqual(km.unit_name, "s^-1")
        self.assertAlmostEqual(np.log(ra.A/km.unit), np.log(3.33e10), 0)
        with tmpdir(__name__, 'test_reaction_analysis_mat1') as dn:
            ra.plot_arrhenius(os.path.join(dn, "arrhenius_mat1.png"))
            ra.monte_carlo()
            ra.write_to_file(os.path.join(dn, "reaction_mat1.txt"))
            ra.plot_parameters(os.path.join(dn, "parameters_mat1.png"))

        wigner = Wigner(pf_ts) # Blind test of the wigner correction and
        # the corrected reaction analysis.
        km = KineticModel([pf_react], pf_ts, tunneling=wigner)
        ra = ReactionAnalysis(km, 100, 1200, temp_step=50)
        with tmpdir(__name__, 'test_reaction_analysis_mat2') as dn:
            ra.plot_arrhenius(os.path.join(dn, "arrhenius_mat1w.png"))
            ra.monte_carlo()
            ra.write_to_file(os.path.join(dn, "reaction_mat1w.txt"))
            ra.plot_parameters(os.path.join(dn, "parameters_mat1w.png"))

        km = KineticModel([pf_react], pf_ts)
        ra = ReactionAnalysis(km, 670, 770)
        # not a very accurate check because the fit is carried out differently
        # in the fancy excel file where these numbers come from.
        self.assertAlmostEqual(ra.Ea/kjmol, 161.9, 1)
        self.assertAlmostEqual(np.log(ra.A/km.unit), np.log(4.08e10), 0)
        with tmpdir(__name__, 'test_reaction_analysis_mat3') as dn:
            ra.plot_arrhenius(os.path.join(dn, "arrhenius_mat2.png"))
            ra.monte_carlo()
            ra.write_to_file(os.path.join(dn, "reaction_mat2.txt"))
            ra.plot_parameters(os.path.join(dn, "parameters_mat2.png"))

    def test_thermo_analysis_mat(self):
        # just a blind test to see test whether the code does not crash.
        pf = PartFun(NMA(load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/mat5T/react.fchk"))),
            [ExtTrans(), ExtRot(1)])
        ta = ThermoAnalysis(pf, [200,300,400,500,600,700,800,900])
        with tmpdir(__name__, 'test_thermo_analysis_mat') as dn:
            ta.write_to_file(os.path.join(dn, "thermo_mat2.csv"))



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


from tamkin.data import Molecule

from molmod import electronvolt, angstrom, amu
from molmod.periodic import periodic
from molmod.unit_cells import UnitCell

import numpy as np


__all__ = ["load_molecule_vasp", "load_fixed_vasp"]


def load_molecule_vasp(contcar, outcar_freq, outcar_energy=None, energy=None, multiplicity=1, is_periodic=True):
    """Load a molecule from VASP 4.6.X and 5.3.X output files

       Arguments:
        | contcar  --  A CONTCAR file with the structure used as POSCAR file for the
                       Hessian/frequency calculation in VASP. Do not use the CONTCAR file
                       generated by the frequency calculation. Use the CONTCAR from the
                       preceding geometry optimization instead.
        | outcar_freq  --  The OUTCAR file of the Hessian/frequency calculation. Also the
                           gradient and the energy are read from this file. The energy
                           without entropy (but not the extrapolation to sigma=0) is used.

       Optional arguments:
        | outcar_energy  --  When given, the (first) energy without entropy is read from
                             this file (not the extrapolation to sigma=0) instead of
                             reading the energy from the freq output
        | energy  --  The potential energy, which overrides the contents of outcar_freq.
        | multiplicity  --  The spin multiplicity of the electronic system
                            [default=1]
        | is_periodic  --  True when the system is periodic in three dimensions.
                           False when the systen is nonperiodic. [default=True].
    """
    # auxiliary function to read energy:
    def read_energy_without_entropy(f):
        # Go to the first energy
        for line in f:
            if line.startswith('  FREE ENERGIE OF THE ION-ELECTRON SYSTEM (eV)'):
                break
        # Skip three lines and read energy
        f.next()
        f.next()
        f.next()
        return float(f.next().split()[3])*electronvolt

    # Read atomic symbols, coordinates and cell vectors from CONTCAR
    symbols = []
    coordinates = []
    with open(contcar) as f:
        # Skip title.
        f.next().strip()

        # Read scale for rvecs.
        rvec_scale = float(f.next())
        # Read rvecs. VASP uses one row per cell vector.
        rvecs = np.fromstring(f.next()+f.next()+f.next(), sep=' ').reshape(3, 3)
        rvecs *= rvec_scale*angstrom
        unit_cell = UnitCell(rvecs)

        # Read symbols
        unique_symbols = f.next().split()
        # Read atom counts per symbol
        symbol_counts = [int(w) for w in f.next().split()]
        assert len(symbol_counts) == len(unique_symbols)
        natom = sum(symbol_counts)
        # Construct array with atomic numbers.
        numbers = []
        for iunique in xrange(len(unique_symbols)):
            number = periodic[unique_symbols[iunique]].number
            numbers.extend([number]*symbol_counts[iunique])
        numbers = np.array(numbers)

        # Check next line
        while f.next() != 'Direct\n':
            continue

        # Load fractional coordinates
        fractional = np.zeros((natom, 3), float)
        for iatom in xrange(natom):
            words = f.next().split()
            fractional[iatom, 0] = float(words[0])
            fractional[iatom, 1] = float(words[1])
            fractional[iatom, 2] = float(words[2])
        coordinates = unit_cell.to_cartesian(fractional)

    if outcar_energy is not None and energy is None:
        with open(outcar_energy) as f:
            energy = read_energy_without_entropy(f)

    # Read energy, gradient, Hessian and masses from outcar_freq. Note that the first
    # energy/force calculation is done on the unperturbed input structure.
    with open(outcar_freq) as f:
        # Loop over the POTCAR sections in the OUTCAR file
        number = None
        masses = np.zeros(natom, float)
        while True:
            line = f.next()
            if line.startswith('   VRHFIN ='):
                symbol = line[11:line.find(':')].strip()
                number = periodic[symbol].number
            elif line.startswith('   POMASS ='):
                mass = float(line[11:line.find(';')])*amu
                masses[numbers==number] = mass
            elif number is not None and line.startswith('------------------------------'):
                assert masses.min() > 0
                break

        # Go to the first gradient
        for line in f:
            if line.startswith(' POSITION'):
                break
        # Skip one line and read the gradient
        f.next()
        gradient = np.zeros((natom, 3), float)
        gunit = electronvolt/angstrom
        for iatom in xrange(natom):
            words = f.next().split()
            gradient[iatom, 0] = -float(words[3])*gunit
            gradient[iatom, 1] = -float(words[4])*gunit
            gradient[iatom, 2] = -float(words[5])*gunit

        if energy is None:
            energy = read_energy_without_entropy(f)

        # Go to the second derivatives
        for line in f:
            if line.startswith(' SECOND DERIVATIVES (NOT SYMMETRIZED)'):  break

        # Skip one line.
        f.next()

        # Load free atoms (not fixed in space).
        keys = f.next().split()
        nfree_dof = len(keys)
        indices_free = [3*int(key[:-1])+{'X': 0, 'Y': 1, 'Z': 2}[key[-1]]-3 for key in keys]
        assert nfree_dof % 3 == 0

        # Load the actual Hessian
        hunit = electronvolt/angstrom**2
        hessian = np.zeros((3*natom, 3*natom), float)
        for ifree0 in xrange(nfree_dof):
            line = f.next()
            irow = indices_free[ifree0]
            # skip first col
            words = line.split()[1:]
            assert len(words) == nfree_dof
            for ifree1 in xrange(nfree_dof):
                icol = indices_free[ifree1]
                hessian[irow, icol] = -float(words[ifree1])*hunit

        # Symmetrize the Hessian
        hessian = 0.5*(hessian + hessian.T)

    return Molecule(
        numbers, coordinates, masses, energy, gradient, hessian,
        multiplicity=multiplicity, periodic=is_periodic, unit_cell=unit_cell)


def load_fixed_vasp(filename):
    """ Load list of fixed atoms from VASP output file

    Argument:
     | filename  --  Filename of VASP output file (OUTCAR)

    VASP can calculate partial Hessians: only a submatrix of the complete Hessian
    is computed with numerical differentiation. The rest of the Hessian elements
    is put to zero. This function determines which atoms have zero rows/cols in
    the Hessian, or, in other words, which were fixed.
    """
    # Read data from out-VASP-file OUTCAR
    f = open(filename)

    # number of atoms (N)
    for line in f:
        if line.strip().startswith("Dimension of arrays:"):  break
    f.next()
    for line in f:
        words = line.split()
        N = int(words[-1])
        break

    # hessian, not symmetrized, useful to find indices of Hessian elements
    for line in f:
        if line.strip().startswith("SECOND DERIVATIVES (NOT SYMMETRIZED)"):  break
    f.next()
    for line in f:
        Nfree = len(line.split())/3   # nb of non-fixed atoms
        break
    # find the non-fixed atoms
    atoms_free = []
    row = 0
    mu = 0
    for line in f:
        if mu==0:
            atom = int(line.split()[0][:-1])
            atoms_free.append(atom-1)
        mu+=1
        row+=1
        if mu >= 3: mu=0
        if row >= 3*Nfree: break
    f.close()

    fixed_atoms = [at for at in xrange(N) if at not in atoms_free]
    return np.array(fixed_atoms)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


# Import the tamkin library
from tamkin import *
# Import unit conversion factors and physical constants
from molmod.units import *
from molmod.constants import *

import sys, os, numpy # standard libraries


# Parse the command line arguments
args = sys.argv[1:]
if len(args) != 1:
    print "One arguments are required: alkane_n"
    sys.exit()

# Hardcoded parameters
temp = 298.15
pressure = 1*atm

# Load the cp2k data and construct a partition function object
molecule = load_molecule_cp2k(
    os.path.join(args[0], "opt.xyz"),
    os.path.join(args[0], "sp/sp.out"),
    os.path.join(args[0], "freq/freq.out"),
    is_periodic=False,
)
nma = NMA(molecule, ConstrainExt())
pf = PartFun(
    nma,
    [ExtTrans(gaslaw=IdealGasLaw(pressure)), ExtRot(), Vibrations(classical=False)],
)


# Write the frequencies to a csv file
f = file(os.path.join(args[0], "freqs.csv"), "w")
print >> f, '"Frequency","Wavenumber","Vibrational temperature"'
print >> f, '"Atomic units","1/cm","K"'
for i in xrange(len(pf.vibrational.freqs)):
    freq = pf.vibrational.freqs[i]
    print >> f, '%e,%f,%f' % (freq, freq/lightspeed*centimeter, 2*numpy.pi*freq/boltzmann)
f.close()


# B) Generate a thermo analysis report
ta = ThermoAnalysis(pf, [100, 200, 300, 400, 500, 600, 700, 800, 900, 100])
ta.write_to_file(os.path.join(args[0], "static_thermo.csv"))



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


import numpy as np


__all__ = [
    "load_chk", "dump_chk",
    "load_indices", "dump_indices",
]



def load_chk(filename):
    """Load a TAMKin checkpoint file

       Argument:
        | filename  --  the file to load from

       The return value is a dictionary whose keys are field labels and the
       values can be None, string, integer, float, boolean or an array of
       strings, integers, booleans or floats.

       The file format is similar to the Gaussian fchk format, but has the extra
       feature that the shapes of the arrays are also stored.
    """
    f = file(filename)
    result = {}
    while True:
        line = f.readline()
        if line == "":
            break
        if len(line) < 54:
            raise IOError("Header lines must be at least 54 characters long.")
        key = line[:40].strip()
        kind = line[47:52].strip()
        value = line[53:-1] # discard newline
        if kind == 'str':
            result[key] = value
        elif kind == 'int':
            result[key] = int(value)
        elif kind == 'bln':
            result[key] = value == "True"
        elif kind == 'flt':
            result[key] = float(value)
        elif kind[3:5] == 'ar':
            if kind[:3] == 'str':
                dtype = str
            elif kind[:3] == 'int':
                dtype = int
            elif kind[:3] == 'bln':
                dtype = bool
            elif kind[:3] == 'flt':
                dtype = float
            else:
                raise IOError("Unsupported kind: %s" % kind)
            shape = tuple(int(i) for i in value.split(","))
            array = np.zeros(shape, dtype)
            work = array.ravel()
            counter = 0
            while True:
                short = f.readline().split()
                for s in short:
                    if dtype is bool and s.lower() in ["True", "1", "Yes"]:
                        work[counter] = True
                    else:
                        work[counter] = dtype(s)
                    counter += 1
                    if counter == array.size:
                        break
                if counter == array.size:
                    break
            result[key] = array
        elif kind == 'none':
            result[key] = None
        else:
            raise IOError("Unsupported kind: %s" % kind)
    f.close()
    return result


def dump_chk(filename, data):
    """Dump a TAMKin checkpoint file

       Argument:
        | filename  --  the file to write to
        | data  -- a dictionary whose keys are field labels and the values can
                   be None, string, integer, float, boolean, an array/list of
                   strings, integers, floats or booleans.

       The file format is similar to the Gaussian fchk format, but has the extra
       feature that the shapes of the arrays are also stored.
    """
    f = file(filename, "w")
    for key, value in sorted(data.iteritems()):
        if not isinstance(key, str):
            raise TypeError("The keys must be strings.")
        if len(key) > 40:
            raise ValueError("Key strings can not be longer than 40 characters.")
        if isinstance(value, str):
            if len(value) > 256:
                raise TypeError("Only small strings are supported (256 chars).")
            if "\n" in value:
                raise ValueError("The string can not contain new lines.")
            print >> f, "%40s  kind=str   %s" % (key.ljust(40), value)
        elif isinstance(value, bool):
            print >> f, "%40s  kind=bln   %s" % (key.ljust(40), value)
        elif isinstance(value, int):
            print >> f, "%40s  kind=int   %i" % (key.ljust(40), value)
        elif isinstance(value, float):
            print >> f, "%40s  kind=flt   %22.15e" % (key.ljust(40), value)
        elif isinstance(value, np.ndarray) or isinstance(value, list) or \
             isinstance(value, tuple):
            if isinstance(value, list) or isinstance(value, tuple):
                value = np.array(value)
            if value.dtype.fields is not None:
                raise TypeError("Arrays with fields are not supported.")
            shape_str = ",".join(str(i) for i in value.shape)
            if issubclass(value.dtype.type, str):
                for cell in value.flat:
                    if len(cell) >= 22:
                        raise ValueError("In case of string arrays, a string may contain at most 21 characters.")
                    if " " in cell or "\n" in cell:
                        raise ValueError("In case of string arrays, a string may not contain spaces or new lines.")
                print >> f, "%40s  kind=strar %s" % (key.ljust(40), shape_str)
                format_str = "%22s"
            elif issubclass(value.dtype.type, int):
                print >> f, "%40s  kind=intar %s" % (key.ljust(40), shape_str)
                format_str = "%22i"
            elif issubclass(value.dtype.type, np.bool_):
                print >> f, "%40s  kind=blnar %s" % (key.ljust(40), shape_str)
                format_str = "%22s"
            elif issubclass(value.dtype.type, float):
                print >> f, "%40s  kind=fltar %s" % (key.ljust(40), shape_str)
                format_str = "%22.15e"
            else:
                raise TypeError("Numpy array dtype %s not supported." % value.dtype)
            short_len = 4
            short = []
            for x in value.ravel():
                short.append(x)
                if len(short) == 4:
                    print >> f, " ".join(format_str  % s for s in short)
                    short = []
            if len(short) > 0:
                print >> f, " ".join(format_str  % s for s in short)
        elif value is None:
            print >> f, "%40s  kind=none   None" % key.ljust(40)
        else:
            raise TypeError("Type %s not supported." % type(value))
    f.close()


def _convert_range(s, shift):
    try:
        return [int(s) + shift]
    except ValueError:
        pass
    if not (s.count(',') == 1 and s.startswith('[') and
            (s.endswith('[') or s.endswith(']'))):
        raise ValueError('Cannot interpret "%s" as a range.' % s)
    b, e = s[1:-1].split(',')
    b = int(b) + shift
    e = int(e) + shift
    result = range(b, e)
    if s.endswith(']'):
        result.append(e)
    return result


def load_indices(filename, shift=None, groups=False):
    """Load atom indexes from file

       Individual atom indices are separated by white space or by one new-line
       character. Ranges of atoms can be specified in two ways:

       * ``[N1,N2]``: the range includes N2 (Fortranish).
       * ``[N1,N2[``: the range does not include N2 (Pythonic).

       The atom indexes can be separated by one or more empty lines
       to define groups of atoms that belong toghether in one block. (See the groups
       argument.)

       Arguments:
         | filename  --  The file to load the indexes from

       Optional arguments:
         | shift  --  A constant shift applied to all atom indexes to convert
                      between numbers starting from zero and numbers starting from
                      one. When not given, the file is searched from a line beginning
                      with ``#shift=``, where the value after this prefix is used as
                      translation to bring the indices back to C convention. Only the
                      first occurence is parsed. If such line is not present, Fortran
                      convention is assumed (shift=-1).
         | groups  --  When True, the function always returns a list of lists,
                       even when only one group of indexes is found in the file.
                       Otherwise only a single list of indexes is returned, even
                       when multiple groups of indexes are encountered.

    """
    # Read all lines
    with open(filename) as f:
        lines = f.readlines()
    # Determine the shift
    if shift is None:
        for line in lines:
            if line.startswith('#shift='):
                shift = int(line[7:])
                break
    if shift is None:
        shift = -1
    # Read the blocks
    blocks = [[]]
    for line in lines:
        words = line[:line.find('#')].split()
        if len(line.strip()) == 0 and len(blocks[-1]) > 0:
            # start new block
            blocks.append([])
        else:
            for word in words:
                blocks[-1].extend(_convert_range(word, shift))
    # Remove last block if empty, may happen when file ends with empty lines.
    if len(blocks[-1]) == 0:
        del blocks[-1]
    # Groups or not?
    if groups:
        return blocks
    else:
        return sum(blocks, [])


def dump_indices(filename, indices, shift=1, compact=True):
    """Dump atom indexes to file

       Arguments:
         | filename  --  the file to dump the atom indexes to

       Optional arguments:
         | indices  --  a list of atom indices or a list of lists of atom indices
                        (the latter is used to define blocks of atoms)
         | shift  --  A constant shift applied to all atom indexes to convert
                      between numbers starting from zero and numbers starting
                      from one.
         | compact  --  When True, the numbers belonging to the same block are
                        put on one line, otherwise they are separated by a
                        newline character. Different blocks are separated by an
                        empty line. [default=True]
    """

    if len(indices) > 0 and not hasattr(indices[0], "__len__"):
        indices = [indices]

    separator = {True: " ", False: "\n"}[compact]

    with open(filename, "w") as f:
        print >> f, "#shift={}".format(-shift)
        for l in indices:
            group_str = separator.join(str(index+shift) for index in l)
            print >> f, group_str
            print >> f
